from textual import work
from textual.app import ComposeResult
from textual.widgets import Static, Label
from textual.reactive import reactive
from textual.containers import  Vertical, Horizontal
from docker import DockerClient
from docker.models.networks import Network

from .custom_widgets import ResponsiveGrid


class NetworkList(ResponsiveGrid):
    networks_count = reactive(0)

    def __init__(self, docker: DockerClient, **kargs):
        self.networks = []
        self.docker = docker
        super().__init__(**kargs)

    def on_mount(self) -> None:
        self.get_networks()
        self.set_interval(2, self.count_timer)

    def count_timer(self) -> None:
        self.get_networks()

    async def watch_networks_count(self, count: int) -> None:
        await self.grid.remove_children()
        for c in self.networks:
            cw = NetworkWidget(c, self.docker)  # type: ignore
            self.grid.mount(cw)

    @work(exclusive=True)
    def get_networks(self) -> None:
        self.networks = self.docker.networks.list()
        self.networks_count = len(self.networks)


class NetworkWidget(Static):
    def __init__(self, network: Network, client: DockerClient, **kargs):
        self.network = network
        self.client = client
        self.attrs = network.attrs or {}
        self.isize = self.attrs.get("Size", 0) / 1000 / 1000
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label("[b]" + (self.network.name or "")),
            Label(self.network.short_id),
            Horizontal(
                Label("Scope: " + self.attrs.get("Scope", "-")),
                Label(" Driver: " + self.attrs.get("Driver", "-")),
            ),
        )



from textual import work
from textual.app import ComposeResult
from textual.widgets import Static, Label
from textual.reactive import reactive
from textual.containers import Vertical, Horizontal
from docker import DockerClient
from docker.models.volumes import Volume

from .custom_widgets import ResponsiveGrid


class VolumesList(ResponsiveGrid):
    volumes_count = reactive(0)

    def __init__(self, docker: DockerClient, **kargs):
        self.volumes = []
        self.docker = docker
        super().__init__(**kargs)

    def on_mount(self) -> None:
        self.get_volumes()
        self.set_interval(2, self.count_timer)

    def count_timer(self) -> None:
        self.get_volumes()

    async def watch_volumes_count(self, count: int) -> None:
        await self.grid.remove_children()
        for c in self.volumes:
            cw = VolumeWidget(c, self.docker)  # type: ignore
            self.grid.mount(cw)

    @work(exclusive=True)
    def get_volumes(self) -> None:
        self.volumes = self.docker.volumes.list()
        self.volumes_count = len(self.volumes)


class VolumeWidget(Static):
    def __init__(self, volume: Volume, client: DockerClient, **kargs):
        self.volume = volume
        self.client = client
        self.attrs = volume.attrs or {}
        self.isize = self.attrs.get("Size", 0) / 1000 / 1000
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label("[b]" + self.volume.name),
            Label(self.attrs.get("Mountpoint", "")),
            Horizontal(
                Label("Scope: " + self.attrs.get("Scope", "-")),
                Label(" Driver: " + self.attrs.get("Driver", "-")),
            ),
        )



from textual import work
from textual.app import App, ComposeResult
from textual.widgets import Footer, Static, ContentSwitcher, Tabs, Tab, Label
from textual.widgets._header import HeaderClock
from textual.containers import (
    VerticalScroll,
    Horizontal,
    Vertical,
    Container as Group,
)
from textual.reactive import reactive
from docker import DockerClient, errors
from docker.models.containers import Container

from .utils import get_cpu_usage, get_mem_usage
from .logs import LogsButton
from .custom_widgets import CustomButton, ResponsiveGrid, ReactiveString
from .images import ImagesList
from .networks import NetworkList
from .volumes import VolumesList


class AppGUI(App):
    CSS_PATH = "style.css"
    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        ("q", "quit", "Quit"),
    ]
    TITLE = "DOCKERY"

    def __init__(self, docker: DockerClient, **kargs):
        self.docker = docker
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        with Horizontal(id="header"):
            yield HeaderClock()
            yield Tabs(
                Tab("Containers", id="container-list"),
                Tab("Images", id="image-list"),
                Tab("Networks", id="network-list"),
                Tab("Volumes", id="volume-list"),
                Tab("Logs", id="container-logs"),
                id="nav",
            )
        yield Footer()
        with ContentSwitcher():
            yield ContainersList(self.docker, id="container-list")
            yield VerticalScroll(id="container-logs")
            yield ImagesList(self.docker, id="image-list")
            yield NetworkList(self.docker, id="network-list")
            yield VolumesList(self.docker, id="volume-list")

    def on_tabs_tab_activated(self, event: Tabs.TabActivated) -> None:
        self.query_one(ContentSwitcher).current = event.tab.id


class ContainersList(ResponsiveGrid):
    container_count = reactive(0)

    def __init__(self, docker: DockerClient, **kargs):
        self.containers = []
        self.docker = docker
        super().__init__(**kargs)

    def on_mount(self) -> None:
        self.get_containers()
        self.set_interval(2, self.count_timer)

    def count_timer(self) -> None:
        self.get_containers()

    async def watch_container_count(self, count: int) -> None:
        await self.grid.remove_children()
        for c in self.containers:
            cw = ContainerWidget(c, self.docker)  # type: ignore
            self.grid.mount(cw)

    @work(exclusive=True)
    def get_containers(self) -> None:
        self.containers = self.docker.containers.list(all=True)
        self.container_count = len(self.containers)


class ContainerWidget(Static):
    def __init__(self, container: Container, client: DockerClient, **kargs):
        self.container = container
        self.client = client
        self.container_id = container.id
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Group(
            Label("[b]" + (self.container.name or ""), classes="container-name"),
            Vertical(
                ReactiveString(id="status"),
                Horizontal(ReactiveString(id="cpu"), ReactiveString(id="mem")),
                classes="stats-text",
            ),
            classes="container-info",
        )

        yield Group(StatusButtons(self.container))

    def on_mount(self):
        self.set_interval(1, self.data_timer)
        self.running = True
        self.update_usage()

    def data_timer(self) -> None:
        self.update_data()

    @work(exclusive=True)
    def update_data(self) -> None:
        try:
            self.container.reload()
        except errors.NotFound:
            return None
        else:
            status = self.container.status.capitalize()
            self.query_one("#status", ReactiveString).text = (
                "[green]" if status == "Running" else "[bright_black]"
            ) + status

    @work
    def update_usage(self) -> None:
        c = self.container
        cpu = self.query_one("#cpu", ReactiveString)
        mem = self.query_one("#mem", ReactiveString)
        for stat in c.stats(stream=True, decode=True):
            if not self.running:  # finish the thread
                return None
            cpu.text = f"CPU: {get_cpu_usage(stat):.1f}"
            mem.text = f"MEM: {get_mem_usage(stat):.1f}"

    def on_unmount(self):
        self.running = False


class StatusButtons(Static):
    def __init__(self, container: Container, **kargs):
        self.container = container
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Horizontal(
            StartStopButton(self.container),
            RestartButton(self.container),
            LogsButton(self.container),
            classes="buttons-container",
        )


class StartStopButton(Static):
    running = reactive(bool)

    def __init__(self, container: Container, **kargs):
        self.container = container
        super().__init__(**kargs)

    def on_click(self) -> None:
        self.running = not self.running
        if self.running:
            self.container.start()
        else:
            self.container.stop()

    def on_mount(self) -> None:
        self.running = self.container.status == "running"

    def watch_running(self, running: bool):
        self.variant = "error" if running else "success"
        text = ":black_square_button:Stop" if running else ":arrow_forward: Start"
        btn = self.query_one(CustomButton)
        btn.text = text
        btn.color = "red" if running else "green"

    def compose(self) -> ComposeResult:
        yield CustomButton("Start")


class RestartButton(Static):
    def __init__(self, container: Container, **kargs):
        self.container = container
        super().__init__(**kargs)

    def on_click(self) -> None:
        self.container.restart()

    def compose(self) -> ComposeResult:
        yield CustomButton(":grey_exclamation:Restart", color="orange")



from textual import work
import time
from textual.app import ComposeResult
from textual.widgets import Static, TextLog, Tabs
from textual.reactive import reactive
from textual.containers import VerticalScroll
from docker.models.containers import Container

from .custom_widgets import CustomButton


class LogsButton(Static):
    def __init__(self, container: Container, **kargs):
        self.container = container
        super().__init__(**kargs)

    async def on_click(self) -> None:
        cl = self.app.query_one("VerticalScroll#container-logs", VerticalScroll)
        await cl.remove_children()
        lc = LogsContainer(self.container)
        await cl.mount(lc)
        self.app.query_one("#nav", Tabs).active = "container-logs"

    def compose(self) -> ComposeResult:
        yield CustomButton(":notebook:Logs", color="blue")


class LogsContainer(TextLog):
    last_log = reactive("")

    def __init__(self, container: Container, **kargs):
        self.container = container
        super().__init__(highlight=True, auto_scroll=True, wrap=True, **kargs)

    def on_mount(self) -> None:
        self.running = True
        self.update_log()

    async def watch_last_log(self, new_log: str):
        self.write(new_log)

    @work
    def update_log(self) -> None:
        # Get the last 40 logs(get all logs can be slow)
        logs: bytes = self.container.logs(tail=40)
        self.last_log = logs.decode("utf-8", errors="ignore")
        # Start streaming logs(since last second)
        for log in self.container.logs(stream=True, since=time.time() - 1):
            if not self.running:
                # Finish the thread after removing the widget
                # TODO: it doesn't finish immediately? fix?
                return None
            self.last_log = log.decode("utf-8", errors="ignore")

    def on_unmount(self):
        self.running = False



import json
from typing import Any, Literal
from rich.console import Console
from rich.syntax import Syntax
import yaml

console = Console()


def get_cpu_usage(stats: dict) -> float:
    cpu_stats = stats["cpu_stats"]
    precpu_stats = stats["precpu_stats"]
    delta = (
        cpu_stats["cpu_usage"]["total_usage"] - precpu_stats["cpu_usage"]["total_usage"]
    )
    system_delta = cpu_stats.get("system_cpu_usage", 0) - precpu_stats.get(
        "system_cpu_usage", 0
    )
    len_cpu = cpu_stats.get("online_cpus", 1)

    if system_delta == 0:
        return 0.0

    percentage = (delta / system_delta) * len_cpu * 100
    return percentage


def get_mem_usage(stats: dict) -> float:
    mem_stats = stats["memory_stats"]
    if len(mem_stats) == 0:
        return 0.0
    mem_used = (
        mem_stats["usage"]
        - mem_stats["stats"].get("cache", 0)
        + mem_stats["stats"]["active_file"]
    )
    limit = stats["memory_stats"]["limit"]
    percentage = mem_used / limit * 100
    return percentage


def var_dump(obj: Any, syntax: Literal["json", "yaml"] = "yaml"):
    if syntax == "yaml":
        text_obj = yaml.safe_dump(obj, indent=2)
    elif syntax == "json":
        text_obj = json.dumps(obj, default=str, indent=2)
    out = Syntax(text_obj, syntax, theme="ansi_dark")
    console.print(out)



import math
from textual.app import ComposeResult
from textual.widgets import Static
from textual.reactive import reactive
from textual.containers import VerticalScroll, Grid
from textual.events import Resize
from rich.text import TextType


class CustomButton(Static):
    text = reactive("")
    color = reactive("")

    def __init__(self, text: str, variant="round", color: str = "green", **kargs):
        self.variant = variant
        self.start_text = text
        self.start_color = color
        super().__init__(shrink=True, expand=True, **kargs)

    def render(self) -> TextType:
        return self.text

    def watch_color(self, new_color: str):
        if new_color != "":
            self.styles.outline = (self.variant, new_color)

    def on_mount(self) -> None:
        self.text = self.start_text
        self.color = self.start_color


class ResponsiveGrid(VerticalScroll):
    container_count = reactive(0)

    def __init__(self, **kargs):
        self.grid = Grid()
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield self.grid

    def on_mount(self) -> None:
        self.resize()
        self.grid.styles.grid_columns = "1fr"
        self.grid.styles.grid_rows = "4"
        self.grid.styles.width = "100%"
        self.grid.styles.height = "auto"

    def on_resize(self, e: Resize):
        self.resize()

    def resize(self):
        min_container_width = 75
        columns = math.floor(self.size.width / min_container_width)
        self.grid.styles.grid_size_columns = columns


class ReactiveString(Static):
    text = reactive("")

    def __init__(self, **kargs):
        super().__init__(shrink=True, expand=True, **kargs)

    def render(self) -> TextType:
        return self.text



import docker
from docker import errors
from docker.models.containers import Container
from docker.models.volumes import Volume
from docker.models.images import Image
from docker.models.networks import Network
import click
from rich import print

from .gui import AppGUI
from .utils import var_dump

default_options = [
    click.option(
        "--server",
        default=None,
        help="Docker server url, for example: unix:///var/run/docker.sock",
    ),
    click.option("--ssh", is_flag=True),
    click.option("--api-version", default="1.35", help="Docker API version"),
    click.option("--format", "-f", default="yaml", help="Output format"),
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def get_client(server, ssh, api_version, **kargs):
    if server is None:
        client = docker.from_env()
    else:
        client = docker.DockerClient(
            base_url=server, version=api_version, use_ssh_client=ssh
        )
    return client


def run_gui(**kargs):
    client = get_client(**kargs)
    gui = AppGUI(client)
    gui.run()


@click.group(invoke_without_command=True)
@add_options(default_options)
@click.pass_context
def main(ctx, **kargs):
    if ctx.invoked_subcommand is None:
        run_gui(**kargs)


@click.command
@add_options(default_options)
def df(**kargs):
    client = get_client(**kargs)
    var_dump(client.df(), kargs["format"])


@click.command
@add_options(default_options)
def volumes(**kargs):
    client = get_client(**kargs)
    vlms: list[Volume] = client.volumes.list()  # type: ignore
    vlm_list = list(map(lambda x: x.attrs, vlms))
    var_dump(vlm_list, kargs["format"])


@click.command
@add_options(default_options)
@click.option("--all", "-a", is_flag=True)
def ps(**kargs):
    client = get_client(**kargs)
    containers: list[Container] = client.containers.list(
        all=kargs["all"]
    )  # type: ignore
    con_list = list(map(lambda x: x.attrs, containers))
    var_dump(con_list, kargs["format"])


@click.command
@add_options(default_options)
def stats(**kargs):
    client = get_client(**kargs)
    containers: list[Container] = client.containers.list(all=True)  # type: ignore
    stats = list(map(lambda x: x.stats(stream=False), containers))
    var_dump(stats, kargs["format"])


@click.command
@add_options(default_options)
def images(**kargs):
    client = get_client(**kargs)
    imgs: list[Image] = client.images.list()  # type: ignore
    img_list = list(map(lambda x: x.attrs, imgs))
    var_dump(img_list, kargs["format"])


@click.command
@add_options(default_options)
def networks(**kargs):
    client = get_client(**kargs)
    netw: list[Network] = client.networks.list()  # type: ignore
    net_list = list(map(lambda x: x.attrs, netw))
    var_dump(net_list, kargs["format"])


@click.command
@add_options(default_options)
@click.argument("container")
@click.option("--stream", is_flag=True)
def logs(**kargs):
    client = get_client(**kargs)
    try:
        container: Container = client.containers.get(kargs["container"])  # type: ignore
    except errors.NotFound:
        print(f"Container [b]{kargs['container']}[/] not found")
    else:
        if kargs["stream"]:
            for log in container.logs(stream=True):
                print(log.decode("utf-8", errors="ignore"), end="")
        else:
            logs: bytes = container.logs()
            print(logs.decode("utf-8", errors="ignore"))


main.add_command(df)
main.add_command(volumes)
main.add_command(ps)
main.add_command(stats)
main.add_command(images)
main.add_command(networks)
main.add_command(logs)

if __name__ == "__main__":
    main()



from textual import work
from textual.app import ComposeResult
from textual.widgets import Static, Label
from textual.reactive import reactive
from textual.containers import Vertical
from docker import DockerClient
from docker.models.images import Image

from .custom_widgets import ResponsiveGrid


class ImagesList(ResponsiveGrid):
    images_count = reactive(0)

    def __init__(self, docker: DockerClient, **kargs):
        self.images = []
        self.docker = docker
        super().__init__(**kargs)

    def on_mount(self) -> None:
        self.get_images()
        self.set_interval(2, self.count_timer)

    def count_timer(self) -> None:
        self.get_images()

    async def watch_images_count(self, count: int) -> None:
        await self.grid.remove_children()
        for c in self.images:
            cw = ImageWidget(c, self.docker)  # type: ignore
            self.grid.mount(cw)

    @work(exclusive=True)
    def get_images(self) -> None:
        self.images = self.docker.images.list(all=False)
        self.images_count = len(self.images)


class ImageWidget(Static):
    def __init__(self, image: Image, client: DockerClient, **kargs):
        self.image = image
        self.client = client
        self.attrs = image.attrs or {}
        self.isize = self.attrs.get("Size", 0) / 1000 / 1000
        if self.image.tags:
            self.tag = self.image.tags[0]
        elif self.image.id:
            self.tag = self.image.id.replace("sha256:", "")
        else:
            self.tag = ""
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label("[b]" + self.tag),
            Label(self.image.short_id.replace("sha256:", "")),
            Label(f"Size: {self.isize:.2f}MB"),
        )
